import logging
import logging.config
import yaml
import traceback
import datetime
import time
import json
import threading
import signal
import sys

import redis
import consul
from flask import Flask
from flask import jsonify

from biomaj_core.config import BiomajConfig
from biomaj.bank import Bank
from biomaj.notify import Notify
from biomaj_core.utils import Utils

from biomaj_zipkin.zipkin import Zipkin


app = Flask(__name__)
app_log = logging.getLogger('werkzeug')
app_log.setLevel(logging.ERROR)


@app.route('/api/daemon-message')
def ping():
    return jsonify({'msg': 'pong'})


def start_web(config):
    app.run(host='0.0.0.0', port=config['web']['port'])


def consul_declare(config):
    if config['consul']['host']:
        consul_agent = consul.Consul(host=config['consul']['host'])
        consul_agent.agent.service.register(
            'biomaj-daemon-message',
            service_id=config['consul']['id'],
            address=config['web']['hostname'],
            port=config['web']['port'],
            tags=['biomaj']
        )
        check = consul.Check.http(
            url='http://' + config['web']['hostname'] + ':' + str(config['web']['port']) + '/api/daemon-message',
            interval=20
        )
        consul_agent.agent.check.register(
            config['consul']['id'] + '_check',
            check=check,
            service_id=config['consul']['id']
        )
        return True
    else:
        return False


class Options(object):
    def __init__(self, d):
        self.__dict__ = d

    def get_option(self, option):
        """
        Gets an option if present, else return None
        """
        if hasattr(self, option):
            return getattr(self, option)
        return None


class DaemonService(object):

    channel = None

    def supervise(self):
        if consul_declare(self.config):
            web_thread = threading.Thread(target=start_web, args=(self.config,))
            web_thread.start()

    def __init__(self, config_file):
        self.logger = logging
        self.curBank = None
        self.session = None
        self.executed_callback = None
        with open(config_file, 'r') as ymlfile:
            self.config = yaml.load(ymlfile)
            Utils.service_config_override(self.config)

        Zipkin.set_config(self.config)

        BiomajConfig.load_config(self.config['biomaj']['config'])
        for svc in Utils.services:
            service = svc.lower()
            if self.config['web'].get('local_endpoint_' + service, None):
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.service.' + service, '1')
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.proxy.' + service, self.config['web']['local_endpoint_' + service])
        if self.config['web'].get('local_endpoint', None):
            BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.proxy', self.config['web']['local_endpoint'])
        if self.config.get('rabbitmq', None):
            if self.config['rabbitmq'].get('host', None):
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.rabbit_mq', self.config['rabbitmq']['host'])
            if self.config['rabbitmq'].get('port', None):
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.rabbit_mq_port', str(self.config['rabbitmq']['port']))
            if self.config['rabbitmq'].get('user', None):
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.rabbit_mq_user', self.config['rabbitmq']['user'])
            if self.config['rabbitmq'].get('password', None):
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.rabbit_mq_password', self.config['rabbitmq']['password'])
            if self.config['rabbitmq'].get('virtual_host', None):
                BiomajConfig.global_config.set('GENERAL', 'micro.biomaj.rabbit_mq_virtual_host', self.config['rabbitmq']['virtual_host'])

        if 'log_config' in self.config:
            for handler in list(self.config['log_config']['handlers'].keys()):
                self.config['log_config']['handlers'][handler] = dict(self.config['log_config']['handlers'][handler])
            logging.config.dictConfig(self.config['log_config'])
            self.logger = logging.getLogger('biomaj')

        self.redis_client = redis.StrictRedis(
            host=self.config['redis']['host'],
            port=self.config['redis']['port'],
            db=self.config['redis']['db'],
            decode_responses=True
        )

        self.logger.info('Daemon service started')
        signal.signal(signal.SIGTERM, self.catch)
        signal.siginterrupt(signal.SIGTERM, False)

    def catch(self, signum, frame):
        self.logger.warn('SIGTERM signal received')
        if self.curBank:
            self.redis_client.set(self.config['redis']['prefix'] + ':' + self.curBank + ':action:cancel', 1)
            self.logger.warn('SIGTERM signal received, cancelling update for ' + self.curBank)
        else:
            sys.exit(1)

    def close(self):
        if self.channel:
            self.channel.close()

    def on_executed_callback(self, func):
        self.executed_callback = func

    def __start_action(self, bank, action):
        self.curBank = bank
        whatsup = bank + ':' + str(action)
        self.redis_client.hset(
            self.config['redis']['prefix'] + ':daemons:status',
            self.config['consul']['id'],
            whatsup
        )
        # Expires in 7 days if no update
        self.redis_client.expire(
            self.config['redis']['prefix'] + ':daemons:status',
            3600 * 24 * 7
        )

    def __end_action(self):
        self.redis_client.hset(
            self.config['redis']['prefix'] + ':daemons:status',
            self.config['consul']['id'],
            'pending'
        )
        self.curBank = None

    def execute(self, options):
        '''
        Execute update or remove command
        '''
        start_time = datetime.datetime.now()
        start_time = time.mktime(start_time.timetuple())

        is_ok = None
        is_updated = False
        action = None
        try:
            options.no_log = False
            if options.update:
                action = 'update'
                self.__start_action(options.bank, action)
                bmaj = Bank(options.bank, options)
                self.logger.debug('Log file: ' + bmaj.config.log_file)
                is_ok = bmaj.update(depends=True)
                is_updated = bmaj.session.get('update')
                Notify.notifyBankAction(bmaj)
                self.__end_action()
            elif options.remove or options.removeall:
                action = 'remove'
                self.__start_action(options.bank, action)
                if options.removeall:
                    bmaj = Bank(options.bank, options, no_log=True)
                    print('Log file: ' + bmaj.config.log_file)
                    is_ok = bmaj.removeAll(options.force)
                else:
                    bmaj = Bank(options.bank, options)
                    self.logger.debug('Log file: ' + bmaj.config.log_file)
                    is_ok = bmaj.remove(options.release)
                    Notify.notifyBankAction(bmaj)
                self.__end_action()
            elif options.removepending:
                bmaj = Bank(options.bank, options, no_log=True)
                bmaj.remove_pending(options.release)
        except Exception as e:
            self.logger.exception('Exception: ' + str(e))
            is_ok = False

        end_time = datetime.datetime.now()
        end_time = time.mktime(end_time.timetuple())

        execution_time = end_time - start_time
        return {
            'error': not is_ok,
            'execution_time': execution_time,
            'action': action,
            'updated': is_updated
        }

    def callback_messages(self, body):
        '''
        Manage bank update or removal
        '''
        try:
            options = Options(body)
            info = self.execute(options)
            if self.executed_callback and options.bank:
                self.executed_callback(options.bank, [info])
        except Exception as e:
            self.logger.error('Error with message: %s' % (str(e)))
            traceback.print_exc()

    def wait_for_messages(self):
        '''
        Loop queue waiting for messages
        '''
        while True:
            msg = self.redis_client.rpop(self.config['redis']['prefix'] + ':queue')
            if msg:
                msg = json.loads(msg)
                info = None
                if 'update' in msg and msg['update']:
                    info = 'update'
                if 'remove' in msg and msg['remove']:
                    info = 'remove'
                if 'removeall' in msg and msg['removeall']:
                    info = 'removeall'
                logging.info('Biomaj:Daemon:' + msg['bank'] + ':' + str(info))

                span = None
                if 'trace' in msg and msg['trace']:
                    span = Zipkin('biomaj', 'workflow')
                    span.add_binary_annotation('operation', info)
                    msg['traceId'] = span.get_trace_id()
                    msg['spanId'] = span.get_span_id()

                self.callback_messages(msg)

                if span:
                    span.trace()

                logging.info('Biomaj:Daemon:' + msg['bank'] + ':over')
            time.sleep(1)



import ssl
import os
import yaml
import logging
from collections import deque
import copy

from flask import Flask
from flask import jsonify
from flask import request
from flask import abort
from flask import Response

import requests

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client.exposition import generate_latest
from prometheus_client import multiprocess
from prometheus_client import CollectorRegistry

import consul
import redis
from pymongo import MongoClient
import pika

from biomaj.schema_version import SchemaVersion

from biomaj.options import Options as BmajOptions
from biomaj_core.config import BiomajConfig
from biomaj_core.utils import Utils
from biomaj.bank import Bank

from biomaj_daemon.daemon.utils import biomaj_client_action

config_file = 'config.yml'
if 'BIOMAJ_CONFIG' in os.environ:
    config_file = os.environ['BIOMAJ_CONFIG']

config = None
with open(config_file, 'r') as ymlfile:
    config = yaml.load(ymlfile)
    Utils.service_config_override(config)

BiomajConfig.load_config(config['biomaj']['config'])

last_status_check = None
last_status = None

data_dir = BiomajConfig.global_config.get('GENERAL', 'data.dir')
if not os.path.exists(data_dir):
    os.makedirs(data_dir)
log_dir = BiomajConfig.global_config.get('GENERAL', 'log.dir')
if not os.path.exists(log_dir):
    os.makedirs(log_dir)
process_dir = BiomajConfig.global_config.get('GENERAL', 'process.dir')
if not os.path.exists(process_dir):
    os.makedirs(process_dir)
cache_dir = BiomajConfig.global_config.get('GENERAL', 'cache.dir')
if not os.path.exists(cache_dir):
    os.makedirs(cache_dir)
lock_dir = BiomajConfig.global_config.get('GENERAL', 'lock.dir')
if not os.path.exists(lock_dir):
    os.makedirs(lock_dir)

redis_client = redis.StrictRedis(
    host=config['redis']['host'],
    port=config['redis']['port'],
    db=config['redis']['db'],
    decode_responses=True
)

logging.info("Check database schema and upgrade if necessary")
SchemaVersion.migrate_pendings()
SchemaVersion.set_version()

app = Flask(__name__)

biomaj_metric = Counter("biomaj_daemon_total", "Bank total update execution.", ['bank', 'action', 'updated'])
biomaj_error_metric = Counter("biomaj_daemon_errors", "Bank total update errors.", ['bank', 'action'])
biomaj_time_metric = Gauge("biomaj_daemon_time", "Bank update execution time in seconds.", ['bank', 'action', 'updated'], multiprocess_mode='all')


def consul_declare(config):
    if config['consul']['host']:
        consul_agent = consul.Consul(host=config['consul']['host'])
        consul_agent.agent.service.register(
            'biomaj-daemon',
            service_id=config['consul']['id'],
            address=config['web']['hostname'],
            port=config['web']['port'],
            tags=[
                'biomaj',
                'api',
                'traefik.backend=biomaj-daemon',
                'traefik.frontend.rule=PathPrefix:/api/daemon',
                'traefik.enable=true',
                'traefik-int.backend=biomaj-daemon',
                'traefik-int.frontend.rule=PathPrefix:/api/daemon',
                'traefik-int.enable=true'
                ]
        )
        check = consul.Check.http(url='http://' + config['web']['hostname'] + ':' + str(config['web']['port']) + '/api/daemon', interval=20)
        consul_agent.agent.check.register(config['consul']['id'] + '_check', check=check, service_id=config['consul']['id'])


consul_declare(config)

OPTIONS_PARAMS = {
    'check': False,
    'config': None,
    'update': False,
    'fromscratch': False,
    'publish': False,
    'unpublish': False,
    'release': None,
    'from_task': None,
    'process': None,
    'log': None,
    'remove': False,
    'removeall': False,
    'removepending': False,
    'status': False,
    'bank': None,
    'owner': None,
    'stop_before': None,
    'stop_after': None,
    'freeze': False,
    'unfreeze': False,
    'force': False,
    'help': False,
    'search': False,
    'formats': None,
    'types': None,
    'query': None,
    'show': False,
    'newbank': None,
    'newdir': None,
    'visibility': None,
    'maintenance': None,
    'version': False,
    'statusko': False,
    'trace': False,
    'whatsup': False,
    'lastlog': None,
    'tail': 100,
    'stats': False,
    'json': False,
    'updatestatus': False,
    'updatecancel': False,
    'aboutme': False,
    'userlogin': None,
    'userpassword': None,
    'proxy': None,
    'schedule': False,
    'history': False,
    'historyLimit': 20
}


class Options(object):
    def __init__(self, d):
        self.__dict__ = d
        for key in list(OPTIONS_PARAMS.keys()):
            if not self.has_option(key):
                setattr(self, key, OPTIONS_PARAMS[key])

    def has_option(self, option):
        if hasattr(self, option):
            return True
        else:
            return False

    def get_option(self, option):
        """
        Gets an option if present, else return None
        """
        if hasattr(self, option):
            return getattr(self, option)
        return None


@app.route('/api/daemon', methods=['GET'])
def ping():
    return jsonify({'msg': 'pong'})


@app.route('/api/daemon/status', methods=['GET'])
def biomaj_status_info():
    '''
    if last_status_check is not None:
        cur = datetime.datetime.now()
        if (cur - last_status_check).total_seconds() < 10:
            return status
    last_status_check = datetime.datetime.now()
    '''
    status = {
        'status': [{'service': 'biomaj-public-proxy', 'status': 1, 'count': 1}],
        'biomaj_services': [],
        'general_services': [{
            'id': 'biomaj-public-proxy',
            'host': '',
            'status': True
        }]
    }

    # Check redis
    logging.debug("Status: check redis")
    redis_ok = False
    try:
        pong = redis_client.ping()
        if pong:
            redis_ok = True
    except Exception as e:
        logging.error('Failed to ping redis: ' + str(e))
    if not redis_ok:
        status['status'].append({'service': 'biomaj-redis', 'status': -1, 'count': 0})
    else:
        status['status'].append({'service': 'biomaj-redis', 'status': 1, 'count': 1})
        status['general_services'].append({
            'id': 'biomaj-redis',
            'host': config['redis']['host'],
            'status': True
        })

    # Check internal proxy
    proxy = Utils.get_service_endpoint(config, 'user')
    r = requests.get(proxy + '/api/user')
    if not r.status_code == 200:
        status['status'].append({'service': 'biomaj-internal-proxy', 'status': -1, 'count': 0})
    else:
        status['status'].append({'service': 'biomaj-internal-proxy', 'status': 1, 'count': 1})
        status['general_services'].append({
            'id': 'biomaj-internal-proxy',
            'host': config['web']['local_endpoint'],
            'status': True
        })

    # Check mongo
    logging.debug("Status: check mongo")
    if 'mongo' in config:
        mongo_ok = False
        try:
            biomaj_mongo = MongoClient(config['mongo']['url'])
            biomaj_mongo.server_info()
            mongo_ok = True
        except Exception as e:
            logging.exception('Failed to connect to mongo:' + str(e))
        if not mongo_ok:
            status['status'].append({'service': 'biomaj-mongo', 'status': -1, 'count': 0})
        else:
            status['status'].append({'service': 'biomaj-mongo', 'status': 1, 'count': 1})
            status['general_services'].append({
                'id': 'biomaj-mongo',
                'host': config['mongo']['url'],
                'status': True
            })

    # Check rabbitmq
    logging.debug("Status: check rabbit")
    if 'rabbitmq' in config and 'host' in config['rabbitmq'] and config['rabbitmq']['host']:
        rabbit_ok = False
        channel = None
        try:
            connection = None
            rabbitmq_port = config['rabbitmq']['port']
            rabbitmq_user = config['rabbitmq']['user']
            rabbitmq_password = config['rabbitmq']['password']
            rabbitmq_vhost = config['rabbitmq']['virtual_host']
            if rabbitmq_user:
                credentials = pika.PlainCredentials(rabbitmq_user, rabbitmq_password)
                connection = pika.BlockingConnection(pika.ConnectionParameters(config['rabbitmq']['host'], rabbitmq_port, rabbitmq_vhost, credentials))
            else:
                connection = pika.BlockingConnection(pika.ConnectionParameters(config['rabbitmq']['host']))
            channel = connection.channel()
            rabbit_ok = True
        except Exception as e:
            logging.error('Rabbitmq connection error: ' + str(e))
        finally:
            if channel:
                channel.close()
        if not rabbit_ok:
            status['status'].append({'service': 'biomaj-rabbitmq', 'status': -1, 'count': 0})
        else:
            status['status'].append({'service': 'biomaj-rabbitmq', 'status': 1, 'count': 1})
            status['general_services'].append({
                'id': 'biomaj-rabbitmq',
                'host': config['rabbitmq']['host'],
                'status': True
            })

    logging.debug("Status: check consul services")
    r = requests.get('http://' + config['consul']['host'] + ':8500/v1/agent/services')
    if not r.status_code == 200:
        status['status'].append({'service': 'biomaj-consul', 'status': -1, 'count': 0})
        # last_status = status
        return status

    status['status'].append({'service': 'biomaj-consul', 'status': 1, 'count': 1})
    status['general_services'].append({
        'id': 'biomaj-consul',
        'host': config['consul']['host'],
        'status': True
    })

    consul_services = r.json()
    services = []
    for consul_service in list(consul_services.keys()):
        current_service = consul_services[consul_service]['Service']
        if current_service in services or current_service == 'consul':
            continue
        else:
            services.append(current_service)
        check_r = requests.get('http://' + config['consul']['host'] + ':8500/v1/health/service/' + consul_services[consul_service]['Service'])
        if not check_r.status_code == 200:
            status['status'].append({'service': 'biomaj-consul', 'status': -1, 'count': 0})
            # last_status = status
            return status
        checks = check_r.json()
        nb_service = 0
        nb_service_ok = 0
        for service_check in checks:
            nb_service += 1
            check_status = True
            for check in service_check['Checks']:
                if check['Status'] != 'passing':
                    check_status = False
                    break
            if check_status:
                nb_service_ok += 1
            status['biomaj_services'].append({
                'id': service_check['Service']['Service'],
                'host': service_check['Service']['Address'],
                'status': check_status
            })
        check_status = -1
        if nb_service == nb_service_ok:
            check_status = 1
        elif nb_service_ok == 0:
            check_status = -1
        else:
            check_status = 0
        status['status'].append({'service': consul_services[consul_service]['Service'], 'count': nb_service, 'status': check_status})

    # Check missing services
    biomaj_services = ['biomaj-watcher', 'biomaj-daemon', 'biomaj-download', 'biomaj-process', 'biomaj-user', 'biomaj-cron', 'biomaj-ftp']
    for biomaj_service in biomaj_services:
        if biomaj_service not in services:
            status['status'].append({'service': biomaj_service, 'count': 0, 'status': -1})

    # last_status = status

    return jsonify(status)


@app.route('/api/daemon/bank/<bank>/log', methods=['GET'])
def biomaj_bank_log(bank):
    return biomaj_bank_log_tail(bank, 0)


@app.route('/api/daemon/bank/<bank>/log/<tail>', methods=['GET'])
def biomaj_bank_log_tail(bank, tail=100):
    apikey = request.headers.get('Authorization')
    token = None

    if apikey:
        bearer = apikey.split()
        if bearer[0] == 'APIKEY':
            token = bearer[1]
    log_file = None
    try:
        user = None
        options_object = Options(OPTIONS_PARAMS)
        if token:
            proxy = Utils.get_service_endpoint(config, 'user')
            r = requests.get(proxy + '/api/user/info/apikey/' + token)
            if not r.status_code == 200:
                abort(404, {'message': 'Invalid API Key or connection issue'})
            user = r.json()['user']
            options_object = Options({'user': user['id']})

        bank_log = Bank(bank, options=options_object, no_log=True)
        if bank_log.bank['properties']['visibility'] != 'public' and not bank_log.is_owner():
            abort(403, {'message': 'not authorized to access this bank'})
        if 'status' not in bank_log.bank or 'log_file' not in bank_log.bank['status'] or not bank_log.bank['status']['log_file']['status']:
            return "No log file available"
        log_file = bank_log.bank['status']['log_file']['status']
    except Exception as e:
        logging.exception(e)
        return "Failed to access log file: " + str(e)
    if not log_file or not os.path.exists(log_file):
        return "Cannot access log file %s" % (str(log_file))

    def generate():
        with open(log_file) as fp:
            tail_l = int(tail)
            if tail_l == 0:
                for line in fp:
                    yield line
            else:
                dq = deque(fp, maxlen=tail_l)
                for line in dq:
                    yield line
        yield "##END_OF_LOG"

    return Response(generate(), mimetype='text/plain')


def daemon_api_auth(request):
    apikey = request.headers.get('Authorization')
    token = None
    options_object = None
    if apikey:
        bearer = apikey.split()
        if bearer[0] == 'APIKEY':
            token = bearer[1]
    if 'X-API-KEY' in request.headers:
        token = request.headers['X-API-KEY']
    try:
        options = copy.deepcopy(OPTIONS_PARAMS)
        options_object = Options(options)
        options_object.json = True
        options_object.token = token
        options_object.user = None
        options_object.redis_host = config['redis']['host']
        options_object.redis_port = config['redis']['port']
        options_object.redis_db = config['redis']['db']
        options_object.redis_prefix = config['redis']['prefix']
        options_object.proxy = True
        user = None
        if token:
            proxy = Utils.get_service_endpoint(config, 'user')
            r = requests.get(proxy + '/api/user/info/apikey/' + token)
            if not r.status_code == 200:
                return (404, options, {'message': 'Invalid API Key or connection issue'})
            user = r.json()['user']
            if user:
                options_object.user = user['id']
    except Exception as e:
        logging.exception(e)
        return (500, options_object, str(e))

    return (200, options_object, None)


@app.route('/api/daemon/aboutme', methods=['POST'])
def biomaj_daemon_aboutme():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)

    options.aboutme = True
    params = request.get_json()
    try:
        options.userlogin = params['login']
        options.userpassword = params['password']
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/maintenance', methods=['POST'])
def biomaj_daemon_maintenance_on():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user or 'admin' not in config['biomaj'] or options.user not in config['biomaj']['admin']:
        abort(401, 'This action requires authentication with api key')
    options.maintenance = 'on'
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/maintenance', methods=['DELETE'])
def biomaj_daemon_maintenance_off():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user or 'admin' not in config['biomaj'] or options.user not in config['biomaj']['admin']:
        abort(401, 'This action requires authentication with api key')
    options.maintenance = 'off'
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/maintenance', methods=['GET'])
def biomaj_daemon_maintenance_status():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.maintenance = 'status'
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/whatsup', methods=['GET'])
def biomaj_daemon_maintenance_whatsup():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.whatsup = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/stats', methods=['GET'])
def biomaj_daemon_stats():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.stats = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/version', methods=['GET'])
def biomaj_daemon_version():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.version = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/history', methods=['GET'])
def biomaj_daemon_history():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)

    options.history = True
    options.historyLimit = request.args.get('limit', 100)
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank', methods=['GET'])
def biomaj_daemon_banks_status():
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.status = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>', methods=['GET'])
def biomaj_daemon_bank_status(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.status = True
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/owns', methods=['GET'])
def biomaj_daemon_bank_isowner(bank):
    '''
    Checks that logged user is admin or owner of the bank
    '''
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    bank_log = Bank(bank, options=options, no_log=True)
    if bank_log.is_owner():
        return jsonify({'is_admin_or_owner': True})
    else:
        return jsonify({'is_admin_or_owner': False})


@app.route('/api/daemon/bank/<bank>/check', methods=['GET'])
def biomaj_daemon_bank_check(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.check = True
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/status', methods=['GET'])
def biomaj_daemon_bank_update_status(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.updatestatus = True
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/status/ko', methods=['GET'])
def biomaj_daemon_bank_status_ko(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.statusko = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/cancel', methods=['PUT'])
def biomaj_daemon_bank_cancel(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.updatecancel = True
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/owner/<owner>', methods=['PUT'])
def biomaj_daemon_bank_upate_owner(bank, owner):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.owner = owner
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/visibility/<visibility>', methods=['PUT'])
def biomaj_daemon_bank_upate_visibility(bank, visibility):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.visibility = visibility
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/name/<name>', methods=['PUT'])
def biomaj_daemon_bank_update_name(bank, name):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.newbank = name
    options.bank = bank
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/move', methods=['PUT'])
def biomaj_daemon_bank_update_directory(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    params = request.get_json()
    options.bank = bank
    try:
        options.newdir = params['path']
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>', methods=['POST'])
def biomaj_daemon_bank_update(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.update = True
    options.bank = bank
    params = request.get_json()
    if params is None:
        params = {}
    if 'publish' in params:
        options.publish = params['publish']
    if 'fromscratch' in params:
        options.fromscratch = params['fromscratch']
    if 'stop_before' in params:
        options.stop_before = params['stop_before']
    if 'stop_after' in params:
        options.stop_after = params['stop_after']
    if 'from_task' in params:
        options.from_task = params['from_task']
    if 'process' in params:
        options.process = params['process']
    if 'release' in params:
        options.release = params['release']
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/publish', methods=['PUT'])
def biomaj_daemon_bank_publish(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.bank = bank
    options.publish = True
    params = request.get_json()
    if params is None:
        params = {}
    if 'release' in params:
        options.release = params['release']
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/publish', methods=['DELETE'])
def biomaj_daemon_bank_unpublish(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.bank = bank
    options.unpublish = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>', methods=['DELETE'])
def biomaj_daemon_bank_remove(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.bank = bank
    params = request.get_json()
    if params is None:
        params = {}
    if 'all' in params and params['all']:
        options.removeall = True
        if 'force' in params and params['force']:
            options.force = params['force']
    elif 'pending' in params and params['pending']:
        options.removepending = True
    else:
        options.remove = True
        if 'release' in params:
            options.release = params['release']
        else:
            abort(400, 'missing release')

    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/freeze/<release>', methods=['PUT'])
def biomaj_daemon_bank_freeze(bank, release):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.bank = bank
    options.release = release
    options.freeze = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/freeze/<release>', methods=['DELETE'])
def biomaj_daemon_bank_unfreeze(bank, release):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.bank = bank
    options.release = release
    options.unfreeze = True
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/search', methods=['GET'])
def biomaj_daemon_bank_search(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    options.search = True
    params = request.get_json()
    if params is None:
        params = {}
    if 'formats' in params and params['formats']:
        options.formats = ','.join(params['formats'])
    if 'types' in params and params['types']:
        options.types = ','.join(params['types'])
    if 'query' in params and params['query']:
        options.query = ','.join(params['query'])
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon/bank/<bank>/show', methods=['GET'])
def biomaj_daemon_bank_show(bank):
    (http_code, options, error) = daemon_api_auth(request)
    if error:
        abort(http_code, error)
    if not options.user:
        abort(401, 'This action requires authentication with api key')
    options.bank = bank
    options.show = True
    params = request.get_json()
    if params is None:
        params = {}
    if 'release' in params:
        options.release = params['release']
    try:
        (res, msg) = biomaj_client_action(options, config)
        if res:
            if isinstance(msg, dict):
                return jsonify(msg)
            else:
                return jsonify({'msg': msg})
    except Exception as e:
        abort(500, str(e))


@app.route('/api/daemon', methods=['POST'])
def biomaj_daemon():
    '''
    Execute a command request (bank update, removal, ...)
    '''
    apikey = request.headers.get('Authorization')
    token = None

    if apikey:
        bearer = apikey.split()
        if bearer[0] == 'APIKEY':
            token = bearer[1]
    try:
        params = request.get_json()
        options = params['options']
        options_object = Options(options)
        options_object.token = token
        options_object.user = None
        options_object.redis_host = config['redis']['host']
        options_object.redis_port = config['redis']['port']
        options_object.redis_db = config['redis']['db']
        options_object.redis_prefix = config['redis']['prefix']

        user = None
        if token:
            proxy = Utils.get_service_endpoint(config, 'user')
            r = requests.get(proxy + '/api/user/info/apikey/' + token)
            if not r.status_code == 200:
                abort(404, {'message': 'Invalid API Key or connection issue'})
            user = r.json()['user']
            if user:
                options_object.user = user['id']

        if options_object.maintenance in ['on', 'off']:
            if not options_object.user or 'admin' not in config['biomaj'] or options_object.user not in config['biomaj']['admin']:
                abort(401, {'message': 'This action requires authentication with api key'})

        if options_object.bank:
            bmaj_options = BmajOptions(options_object)
            BiomajConfig(options_object.bank, bmaj_options)

            if not options_object.search and not options_object.show and not options_object.check and not options_object.status:
                if not user:
                    abort(401, {'message': 'This action requires authentication with api key'})

        (res, msg) = biomaj_client_action(options_object, config)
    except Exception as e:
        logging.exception(e)
        return jsonify({'status': False, 'msg': str(e)})
    return jsonify({'status': res, 'msg': msg})


@app.route('/metrics', methods=['GET'])
def metrics():
    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    return generate_latest(registry)


@app.route('/api/daemon/metrics', methods=['POST'])
def add_metrics():
    '''
    Expects a JSON request with an array of {'bank': 'bank_name', 'error': 'error_message', 'execution_time': seconds_to_execute, 'action': 'update|remove', 'updated': 0|1}
    '''
    procs = request.get_json()
    for proc in procs:
        if 'error' in proc and proc['error']:
            biomaj_error_metric.labels(proc['bank'], proc['action']).inc()
        else:
            biomaj_metric.labels(proc['bank'], proc['action'], proc['updated']).inc()
            biomaj_time_metric.labels(proc['bank'], proc['action'], proc['updated']).set(proc['execution_time'])
    return jsonify({'msg': 'OK'})


if __name__ == "__main__":
    context = None
    if config['tls']['cert']:
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        context.load_cert_chain(config['tls']['cert'], config['tls']['key'])
    app.run(host='0.0.0.0', port=config['web']['port'], ssl_context=context, threaded=True, debug=config['web']['debug'])



import os
import pkg_resources
import shutil
import json
import datetime
import time
import redis
import sys
import logging

import requests

from tabulate import tabulate

from biomaj.bank import Bank
from biomaj_core.config import BiomajConfig
from biomaj_core.utils import Utils
from biomaj.workflow import Workflow
from biomaj.workflow import UpdateWorkflow
from biomaj.workflow import RemoveWorkflow
from biomaj.notify import Notify

if sys.version < '3':
    import ConfigParser as configparser
else:
    import configparser


def biomaj_version(options, config):
    '''
    Get biomaj version
    '''
    version = pkg_resources.require('biomaj')[0].version
    # tools = []
    biomaj_modules = [
        'biomaj',
        'biomaj-core',
        'biomaj-daemon',
        'biomaj-release',
        'biomaj-download',
        'biomaj-process',
        'biomaj-cron',
        'biomaj-ftp',
        'biomajwatcher'
        ]

    results = []
    if not options.json:
        results = [["Module", "Release", "Latest"]]

    msg = 'BioMAJ modules version\n'
    for module_name in biomaj_modules:
        (version, latest) = Utils.get_module_version(module_name)
        if version is not None:
            results.append([module_name, version, latest])

    if options.json:
        return (True, {'version': results})

    msg += tabulate(results, headers="firstrow", tablefmt="grid")
    return (True, 'Version: ' + str(msg))


def check_options(options, config):
    if options.stop_after or options.stop_before or options.from_task:
        available_steps = []
        for flow in UpdateWorkflow.FLOW:
            available_steps.append(flow['name'])
        for flow in RemoveWorkflow.FLOW:
            available_steps.append(flow['name'])
        if options.stop_after:
            if options.stop_after not in available_steps:
                return (False, 'Invalid step: ' + options.stop_after)
        if options.stop_before:
            if options.stop_before not in available_steps:
                return (False, 'Invalid step: ' + options.stop_before)
        if options.from_task:
            if options.from_task not in available_steps:
                return (False, 'Invalid step: ' + options.from_task)
    return (True, None)


def biomaj_maintenance(options, config):
    '''
    Maintenance management
    '''
    if options.maintenance not in ['on', 'off', 'status']:
        # print("Wrong maintenance value [on,off,status]")
        return (False, "Wrong maintenance value [on,off,status]")

    data_dir = BiomajConfig.global_config.get('GENERAL', 'data.dir')
    if BiomajConfig.global_config.has_option('GENERAL', 'lock.dir'):
        lock_dir = BiomajConfig.global_config.get('GENERAL', 'lock.dir')
    else:
        lock_dir = data_dir

    maintenance_lock_file = os.path.join(lock_dir, 'biomaj.lock')
    if options.maintenance == 'status':
        if os.path.exists(maintenance_lock_file):
            return (True, "Maintenance: On")
        else:
            return (True, "Maintenance: Off")

    if options.maintenance == 'on':
        f = open(maintenance_lock_file, 'w')
        f.write('1')
        f.close()
        return (True, "Maintenance set to On")

    if options.maintenance == 'off':
        if os.path.exists(maintenance_lock_file):
            os.remove(maintenance_lock_file)
        return (True, "Maintenance set to Off")


def biomaj_owner(options, config):
    '''
    Bank ownership management
    '''
    if not options.bank:
        return (False, "Bank option is missing")
    bank = Bank(options.bank, options=options, no_log=True)
    bank.set_owner(options.owner)
    return (True, None)


def biomaj_visibility(options, config):
    '''
    Bank visibility management
    '''
    if not options.bank:
        return (False, "Bank option is missing")
    if options.visibility not in ['public', 'private']:
        return (False, "Valid values are public|private")

    bank = Bank(options.bank, options=options, no_log=True)
    bank.set_visibility(options.visibility)
    return (True, "Do not forget to update accordingly the visibility.default parameter in the configuration file")


def biomaj_move_production_directories(options, config):
    '''
    Change bank production directories
    '''
    if not options.bank:
        return (False, "Bank option is missing")

    if not os.path.exists(options.newdir):
        return (False, "Destination directory does not exists")

    bank = Bank(options.bank, options=options, no_log=True)
    if not bank.bank['production']:
        return (False, "Nothing to move, no production directory")

    bank.load_session(Workflow.FLOW, None)
    w = Workflow(bank)
    res = w.wf_init()
    if not res:
        return (False, 'Bank initialization failure')

    for prod in bank.bank['production']:
        session = bank.get_session_from_release(prod['release'])
        bank.load_session(Workflow.FLOW, session)
        prod_path = bank.session.get_full_release_directory()
        if os.path.exists(prod_path):
            shutil.move(prod_path, options.newdir)
        prod['data_dir'] = options.newdir
    bank.banks.update({'name': options.bank}, {'$set': {'production': bank.bank['production']}})
    w.wf_over()
    return (True, "Bank production directories moved to " + options.newdir + "\nWARNING: do not forget to update accordingly the data.dir and dir.version properties")


def biomaj_newbank(options, config):
    '''
    Rename a bank
    '''
    if not options.bank:
        return (False, "Bank option is missing")

    bank = Bank(options.bank, options=options, no_log=True)
    conf_dir = BiomajConfig.global_config.get('GENERAL', 'conf.dir')
    bank_prop_file = os.path.join(conf_dir, options.bank + '.properties')
    config_bank = configparser.SafeConfigParser()
    config_bank.read([os.path.join(conf_dir, options.bank + '.properties')])
    config_bank.set('GENERAL', 'db.name', options.newbank)
    newbank_prop_file = open(os.path.join(conf_dir, options.newbank + '.properties'), 'w')
    config_bank.write(newbank_prop_file)
    newbank_prop_file.close()
    bank.banks.update({'name': options.bank}, {'$set': {'name': options.newbank}})
    os.remove(bank_prop_file)
    return (True, "Bank " + options.bank + " renamed to " + options.newbank)


def biomaj_search(options, config):
    '''
    Search within banks
    '''
    msg = ''
    if options.query:
        res = Bank.searchindex(options.query)
        msg += 'Query matches for :' + options.query + '\n'
        results = []
        headers = ["Release", "Format(s)", "Type(s)", "Files"]
        if not options.json:
            results.append(headers)
        for match in res:
            results.append([match['_source']['release'],
                            str(match['_source']['format']),
                            str(match['_source']['types']),
                            ','.join(match['_source']['files'])])
        if options.json:
            return (True, {'matches': results, 'headers': headers})

        msg += tabulate(results, headers="firstrow", tablefmt="grid")
        return (True, msg)
    else:
        formats = []
        if options.formats:
            formats = options.formats.split(',')
        types = []
        if options.types:
            types = options.types.split(',')
        msg += "Search by formats=" + str(formats) + ", types=" + str(types) + '\n'
        res = Bank.search(formats, types, False)
        results = []
        headers = ["Name", "Release", "Format(s)", "Type(s)", 'Published']
        if not options.json:
            results.append(headers)
        for bank in sorted(res, key=lambda bank: (bank['name'])):
            b = bank['name']
            bank['production'].sort(key=lambda n: n['release'], reverse=True)
            for prod in bank['production']:
                iscurrent = ""
                if prod['session'] == bank['current']:
                    iscurrent = "yes"
                if options.json:
                    results.append(
                        [
                            b if b else '',
                            prod['release'],
                            prod['formats'],
                            prod['types'], iscurrent
                        ]
                    )
                else:
                    results.append(
                        [
                            b if b else '',
                            prod['release'],
                            ','.join(prod['formats']),
                            ','.join(prod['types']), iscurrent
                        ]
                    )
        if options.json:
            return (True, {'matches': results, 'headers': headers})

        msg += tabulate(results, headers="firstrow", tablefmt="grid")
        return (True, msg)


def biomaj_show(options, config):
    '''
    show bank details
    '''
    if not options.bank:
        return (False, "Bank option is required")

    bank = Bank(options.bank, options=options, no_log=True)
    results = []
    headers = ["Name", "Release", "Format(s)", "Type(s)", "Tag(s)", "File(s)"]
    if not options.json:
        results.append(headers)
    current = None
    fformat = None
    if 'current' in bank.bank and bank.bank['current']:
        current = bank.bank['current']
    for prod in bank.bank['production']:
        include = True
        release = prod['release']
        if current == prod['session']:
            release += ' (current)'
        if options.release and (prod['release'] != options.release and prod['prod_dir'] != options.release):
            include = False
        if include:
            session = bank.get_session_from_release(prod['release'])
            formats = session['formats']
            afiles = []
            atags = []
            atypes = []
            for fformat in list(formats.keys()):
                for elt in formats[fformat]:
                    if options.json:
                        atypes.append(elt['types'])
                    else:
                        atypes.append(','.join(elt['types']))
                    for tag in list(elt['tags'].keys()):
                        atags.append(elt['tags'][tag])
                    for eltfile in elt['files']:
                        afiles.append(eltfile)
            if options.json:
                results.append([
                    bank.bank['name'],
                    release,
                    fformat,
                    atypes,
                    atags,
                    afiles])
                msg = {'details': results, 'headers': headers}
            else:
                results.append([
                    bank.bank['name'],
                    release,
                    fformat,
                    ','.join(atypes),
                    ','.join(atags),
                    ','.join(afiles)])
                msg = tabulate(results, headers="firstrow", tablefmt="grid")
    return (True, msg)


def biomaj_check(options, config):
    '''
    Check bank properties
    '''
    if not options.bank:
        return (False, "Bank name is missing")

    bank = Bank(options.bank, options=options, no_log=True)
    msg = options.bank + " check: " + str(bank.check()) + "\n"
    return (True, msg)


def biomaj_status(options, config):
    '''
    Get bank status information
    '''
    msg = ''
    if options.bank:
        bank = Bank(options.bank, options=options, no_log=True)
        if bank.bank['properties']['visibility'] != 'public' and not bank.is_owner():
            return (False, 'Access forbidden')
        info = bank.get_bank_release_info(full=True)
        if options.json:
            bank_info = []
            headers = []
            if info['info'] and len(info['info']) > 1:
                headers = info['info'][0]
                bank_info = info['info'][1:]
            prod_info = []
            prod_headers = []
            if info['prod'] and len(info['prod']) > 1:
                prod_headers = info['prod'][0]
                prod_info = info['prod'][1:]
            msg = {
                'bank': {
                    'info': {'headers': headers, 'details': bank_info},
                    'production': {'headers': prod_headers, 'details': prod_info},
                    'pending': []
                    }
                }
        else:
            msg += tabulate(info['info'], headers='firstrow', tablefmt='psql') + '\n'
            msg += tabulate(info['prod'], headers='firstrow', tablefmt='psql') + '\n'
        # do we have some pending release(s)
        if 'pend' in info and len(info['pend']) > 1:
            if options.json:
                msg['bank']['pending'] = info['pend']
            else:
                msg += tabulate(info['pend'], headers='firstrow', tablefmt='psql') + '\n'
    else:
        banks = Bank.list()
        # Headers of output table
        banks_list = []
        headers = ["Name", "Type(s)", "Release", "Visibility", "Last update"]
        if not options.json:
            banks_list.append(headers)
        for bank in sorted(banks, key=lambda k: k['name']):
            try:
                bank = Bank(bank['name'], options=options, no_log=True)
                if bank.bank['properties']['visibility'] == 'public' or bank.is_owner():
                    banks_list.append(bank.get_bank_release_info()['info'])
            except Exception as e:
                logging.error('Failed to load bank %s: %s' % (bank['name'], str(e)))
        if options.json:
            msg = {'headers': headers, 'banks': banks_list}
        else:
            msg += tabulate(banks_list, headers="firstrow", tablefmt="psql") + '\n'
    return (True, msg)


def biomaj_status_ko(options, config):
    '''
    Get failed banks
    '''
    banks = Bank.list()
    banks_list = []
    headers = ["Name", "Type(s)", "Release", "Visibility", "Last run"]
    if not options.json:
        banks_list.append(headers)

    for bank in sorted(banks, key=lambda k: k['name']):
        try:
            bank = Bank(bank['name'], options=options, no_log=True)
            bank.load_session(UpdateWorkflow.FLOW)
            if bank.session is not None:
                if bank.use_last_session and not bank.session.get_status(Workflow.FLOW_OVER):
                    wf_status = bank.session.get('workflow_status')
                    if wf_status is None or not wf_status:
                        banks_list.append(bank.get_bank_release_info()['info'])
        except Exception as e:
            return (False, str(e))
    if options.json:
        msg = {'headers': headers, 'banks': banks_list}
        return (True, msg)
    return (True, tabulate(banks_list, headers="firstrow", tablefmt="psql"))


def biomaj_bank_update_request(options, config):
    '''
    Send bank update request to rabbitmq
    '''
    return biomaj_send_message(options, config)


def biomaj_whatsup(options, config):
    redis_client = None
    whatsup = []

    if options.proxy:
        redis_client = redis.StrictRedis(
            host=config['redis']['host'],
            port=config['redis']['port'],
            db=config['redis']['db'],
            decode_responses=True
        )
        if not redis_client:
            return (False, 'Redis not configured')

        pending_len = redis_client.llen(config['redis']['prefix'] + ':queue')
        if not pending_len:
            pending_len = 0

        whatsup.append(['queue', str(pending_len), 'waiting'])
        for index in range(pending_len):
            pending_action = redis_client.lindex(config['redis']['prefix'] + ':queue', index)
            if pending_action:
                pending_bank = json.loads(pending_action)
                if pending_bank.get('bank', None):
                    whatsup.append(['queued', pending_bank['bank'], 'waiting'])

    if not options.proxy:
        data_dir = BiomajConfig.global_config.get('GENERAL', 'data.dir')
        lock_dir = data_dir
        if BiomajConfig.global_config.has_option('GENERAL', 'lock.dir'):
            lock_dir = BiomajConfig.global_config.get('GENERAL', 'lock.dir')
        lock_files = [f for f in os.listdir(lock_dir) if os.path.isfile(os.path.join(lock_dir, f))]
        for lock_file in lock_files:
            bank_name = lock_file.replace('.lock', '')
            whatsup.append(['system', str(bank_name), 'running'])
        if not whatsup:
            whatsup.append(['system', '', 'pending'])
    else:
        daemons_status = redis_client.hgetall(config['redis']['prefix'] + ':daemons:status')
        msg = 'All daemons pending'
        for daemon in list(daemons_status.keys()):
            if daemons_status[daemon]:
                # bank:action
                proc = daemons_status[daemon].split(':')
                if len(proc) == 2:
                    whatsup.append([daemon] + proc)
                else:
                    whatsup.append([daemon, ''] + proc)
    if whatsup:
        if options.json:
            msg = {'services': whatsup, 'headers': ['daemon', 'bank', 'action']}
        else:
            msg = tabulate(whatsup, ['daemon', 'bank', 'action'], tablefmt="simple")
    return (True, msg)


def biomaj_send_message(options, config):
    '''
    Send message to listener
    '''
    if not options.proxy:
        return (False, 'option not allowed without --proxy option')
    redis_client = redis.StrictRedis(
        host=config['redis']['host'],
        port=config['redis']['port'],
        db=config['redis']['db'],
        decode_responses=True
    )
    if not redis_client:
        return (False, 'Redis not configured')
    cur = datetime.datetime.now()
    options.timestamp = time.mktime(cur.timetuple())
    redis_client.lpush(config['redis']['prefix'] + ':queue', json.dumps(options.__dict__))
    return (True, None)


def biomaj_bank_update(options, config):
    '''
    Update a bank
    '''
    if not options.bank:
        return (False, "Bank name is missing")
    banks = options.bank.split(',')
    gres = True
    msg = ''
    for bank in banks:
        options.bank = bank
        no_log = True
        if not options.proxy:
            no_log = False
        # logging.debug('Options: '+str(options.__dict__))
        bmaj = Bank(bank, options=options, no_log=no_log)
        if bmaj.is_locked():
            return (False, 'Bank is locked due to an other action')
        check_status = bmaj.check()
        if not check_status:
            msg += 'Skip bank ' + options.bank + ': wrong config\n'
            gres = False
            continue
        else:
            msg += 'Bank update request sent for ' + options.bank + '\n'
            if not options.proxy:
                res = bmaj.update(depends=True)
                Notify.notifyBankAction(bmaj)
            else:
                res = biomaj_bank_update_request(options, config)
            if not res:
                msg += 'Failed to send update request for ' + options.bank + '\n'
                gres = False

    if not gres:
        return (False, msg)
    return (True, msg)


def biomaj_freeze(options, config):
    '''
    freeze a bank
    '''
    if not options.bank:
        return (False, "Bank name is missing")
    if not options.release:
        return (False, "Bank release is missing")
    bmaj = Bank(options.bank, options=options)
    res = bmaj.freeze(options.release)
    if not res:
        return (False, 'Failed to freeze the bank release')
    return (True, None)


def biomaj_unfreeze(options, config):
    '''
    unfreeze a bank
    '''
    if not options.bank:
        return (False, "Bank name is missing")
    if not options.release:
        return (False, "Bank release is missing")

    bmaj = Bank(options.bank, options=options)
    res = bmaj.unfreeze(options.release)
    if not res:
        return (False, 'Failed to unfreeze the bank release')
    return (True, None)


def biomaj_remove_request(options, config):
    '''
    Send remove request to rabbitmq
    '''
    return biomaj_send_message(options, config)


def biomaj_remove(options, config):
    '''
    Remove a bank
    '''
    if not options.bank:
        return (False, "Bank name is missing")

    if options.remove and not options.release:
        return (False, "Bank release is missing")

    no_log = True
    if not options.proxy:
        no_log = False
    # If removeall, do not set logs as log dir will be deleted
    if options.removeall:
        no_log = True
    bmaj = Bank(options.bank, options=options, no_log=no_log)
    if bmaj.is_locked():
        return (False, 'Bank is locked due to an other action')

    res = True
    if options.removeall:
        if not options.proxy:
            res = bmaj.removeAll(options.force)
            return (res, '')
        res = biomaj_remove_request(options, config)
    else:
        if not options.proxy:
            res = bmaj.remove(options.release)
            Notify.notifyBankAction(bmaj)
            return (res, '')
        res = biomaj_remove_request(options, config)
    if not res:
        return (False, 'Failed to send removal request')
    return (True, 'Bank removal request sent')


def biomaj_remove_pending_request(options, config):
    '''
    Send remove pending request to rabbitmq
    '''
    return biomaj_send_message(options, config)


def biomaj_remove_pending(options, config):
    '''
    Remove pending releases
    '''
    if not options.bank:
        return (False, "Bank name is missing")
    bmaj = Bank(options.bank, options=options, no_log=True)
    if bmaj.is_locked():
        return (False, 'Bank is locked due to an other action')
    if not options.proxy:
        res = bmaj.remove_pending(options.release)
        return (res, '')
    res = biomaj_remove_pending_request(options, config)
    if not res:
        return (False, 'Failed to send removal request')
    return (True, 'Request sent')


def biomaj_unpublish(options, config):
    '''
    Unpublish a bank
    '''
    if not options.bank:
        return (False, "Bank name is missing")

    bmaj = Bank(options.bank, options=options, no_log=True)
    bmaj.load_session()
    bmaj.unpublish()
    return (True, None)


def biomaj_publish(options, config):
    '''
    Publish a bank
    '''
    if not options.bank:
        return (False, "Bank name or release is missing")
    bmaj = Bank(options.bank, options=options, no_log=True)
    bmaj.load_session()
    bank = bmaj.bank
    session = None
    if options.get_option('release') is None:
        # Get latest prod release
        if len(bank['production']) > 0:
            prod = bank['production'][len(bank['production']) - 1]
            for s in bank['sessions']:
                if s['id'] == prod['session']:
                    session = s
                    break
    else:
        # Search production release matching release
        for prod in bank['production']:
            if prod['release'] == options.release or prod['prod_dir'] == options.release:
                # Search session related to this production release
                for s in bank['sessions']:
                    if s['id'] == prod['session']:
                        session = s
                        break
                break
    if session is None:
        return (False, "No production session could be found for this release")
    bmaj.session._session = session
    bmaj.publish()
    return (True, None)


def biomaj_update_cancel(options, config):
    '''
    Cancel current update of a Bank

    Running actions (download, process) will continue on remote services but will not manage next actions.
    Biomaj process will exit when ready with a *False* status.
    '''
    if not options.proxy:
        return (False, 'option not allowed without --proxy option')
    redis_client = redis.StrictRedis(
        host=config['redis']['host'],
        port=config['redis']['port'],
        db=config['redis']['db'],
        decode_responses=True
    )
    if not redis_client:
        return (False, 'Redis not configured')
    if not options.bank:
        return (False, "Bank name is missing")
    redis_client.set(config['redis']['prefix'] + ':' + options.bank + ':action:cancel', 1)
    return (True, 'Requested to cancel update of bank ' + options.bank + ', update will stop once current actions are over')


def biomaj_update_status(options, config):
    '''
    get the status of a bank during an update cycle
    '''
    if not options.proxy:
        return (False, 'option not allowed without --proxy option')
    redis_client = redis.StrictRedis(
        host=config['redis']['host'],
        port=config['redis']['port'],
        db=config['redis']['db'],
        decode_responses=True
    )
    if not redis_client:
        return (False, 'Redis not configured')
    pending = redis_client.llen(config['redis']['prefix'] + ':queue')
    pending_actions = []
    pending_headers = ['Pending actions', 'Date']
    if not options.json:
        pending_actions.append(pending_headers)
    for index in range(pending):
        pending_action = redis_client.lindex(config['redis']['prefix'] + ':queue', index)
        if pending_action:
            pending_bank = json.loads(pending_action)
            action_time = datetime.datetime.utcfromtimestamp(pending_bank['timestamp'])
            if pending_bank['bank'] == options.bank:
                if pending_bank['update']:
                    pending_actions.append(['Update', action_time])
                elif pending_bank['remove'] or pending_bank['removeall']:
                    pending_actions.append(['Removal - release ' + str(pending_bank['release']), action_time])

    bmaj = Bank(options.bank, options=options, no_log=True)
    if 'status' not in bmaj.bank:
        return (True, 'No update information available')
    status_info = bmaj.bank['status']

    msg = []
    headers = ["Workflow step", "Status"]
    if not options.json:
        msg.append(headers)

    if 'log_file' in status_info:
        msg.append(['log_file', str(status_info['log_file']['status'])])
    if 'session' in status_info:
        msg.append(['session', str(status_info['session'])])
    for flow in UpdateWorkflow.FLOW:
        step = flow['name']
        if step in status_info:
            if status_info[step]['status'] is True:
                msg.append([step, 'over'])
            elif status_info[step]['status'] is False:
                msg.append([step, 'error'])
            else:
                if status_info[Workflow.FLOW_OVER]['status'] is True:
                    msg.append([step, 'skipped'])
                else:
                    msg.append([step, 'waiting'])
        if step in ['postprocess', 'preprocess', 'removeprocess']:
            if status_info.get(step, None) and status_info[step].get('progress', None):
                progress = status_info[step]['progress']
                for proc in list(progress.keys()):
                    msg.append([proc, str(progress[proc])])
    if options.json:
        return (True, {'pending': {'headers': pending_headers, 'actions': pending_actions}, 'status': {'headers': headers, 'process': msg}})
    else:
        return (True, tabulate(pending_actions, headers="firstrow", tablefmt="grid") + tabulate(msg, headers="firstrow", tablefmt="grid"))


def biomaj_user_info(options, config):
    '''
    Get user info, need login/password
    '''
    if not options.userlogin or not options.userpassword:
        return (False, 'Missing login or password')
    if not options.proxy:
        return (False, 'Missing proxy information')
    bindinfo = {'type': 'password', 'value': options.userpassword}
    try:
        r = requests.post(config['web']['local_endpoint'] + '/api/user/bind/user/' + options.userlogin, json=bindinfo)
        if not r.status_code == 200:
            return (False, 'Invalid credentials')
        user = r.json()['user']
    except Exception as e:
        return (False, 'Connection error to proxy: ' + str(e))
    msg = None
    if options.json:
        msg = {
            'user': str(user['id']),
            'email': str(user['email']),
            'apikey': str(user['apikey'])
        }
    else:
        msg = 'User: ' + str(user['id']) + '\n'
        msg += 'Email: ' + str(user['email']) + '\n'
        msg += 'Api key: ' + str(user['apikey']) + '\n'
    return (True, msg)


def biomaj_stats(options, config):
    disk_stats = Bank.get_banks_disk_usage()
    results = []
    headers = ["Bank", "Release", "Size"]
    if not options.json:
        results.append(headers)
    msg = 'BioMAJ statistics\n'
    for disk_stat in disk_stats:
        results.append([disk_stat['name'],
                        'All',
                        str(disk_stat['size'])])
        for rel in disk_stat['releases']:
            results.append(['',
                            rel['name'],
                            str(rel['size'])])
    if options.json:
        msg = {'headers': headers, 'stats': results}
    else:
        msg += tabulate(results, headers="firstrow", tablefmt="grid")
    return (True, msg)


def biomaj_history(options, config):
    history = Bank.get_history(options.historyLimit)
    results = []
    headers = ["Bank", "Action", "Start", "End", "Updated", "Error"]
    if not options.json:
        results.append(headers)
    msg = 'BioMAJ history\n'
    for h in history:
        results.append(
            [
                h['bank'],
                h['action'],
                datetime.datetime.utcfromtimestamp(h['start']),
                datetime.datetime.utcfromtimestamp(h['end']),
                str(h['updated']),
                str(h['error'])
            ]
        )
    if options.json:
        msg = {'headers': headers, 'history': results}
    else:
        msg += tabulate(results, headers="firstrow", tablefmt="grid")
    return (True, msg)


def biomaj_client_action(options, config=None):
    check_options(options, config)
    if options.version:
        return biomaj_version(options, config)

    if options.history:
        return biomaj_history(options, config)

    if options.stats:
        return biomaj_stats(options, config)

    if options.whatsup:
        return biomaj_whatsup(options, config)

    if options.maintenance:
        return biomaj_maintenance(options, config)

    if options.owner:
        return biomaj_owner(options, config)

    if options.visibility:
        return biomaj_visibility(options, config)

    if options.newdir:
        return biomaj_move_production_directories(options, config)

    if options.newbank:
        return biomaj_newbank(options, config)

    if options.search:
        return biomaj_search(options, config)

    if options.show:
        return biomaj_show(options, config)

    if options.check:
        return biomaj_check(options, config)

    if options.status:
        return biomaj_status(options, config)

    if options.statusko:
        return biomaj_status_ko(options, config)

    if options.update:
        return biomaj_bank_update(options, config)

    if options.freeze:
        return biomaj_freeze(options, config)

    if options.unfreeze:
        return biomaj_unfreeze(options, config)

    if options.remove or options.removeall:
        return biomaj_remove(options, config)

    if options.removepending:
        return biomaj_remove_pending(options, config)

    if options.unpublish:
        return biomaj_unpublish(options, config)

    if options.publish:
        return biomaj_publish(options, config)

    if options.updatestatus:
        return biomaj_update_status(options, config)

    if options.updatecancel:
        return biomaj_update_cancel(options, config)

    if options.aboutme:
        return biomaj_user_info(options, config)

    return (False, "no option match")
